/*
 *  triplex.c: command-line application section of the triplex search
 *             program. Search for potential triplex sequences in DNA.
 *  Copyright (C) 2011
 *    Brno University of Technology, Faculty of Information Technology
 *    Masaryk University, Faculty of Informatics
 *
 *  Author(s): Tomas Martinek <martinto@fit.vutbr.cz>
 *             Daniel Kopecek <xkopecek@fi.muni.cz>
 *             Matej Lexa <lexa@fi.muni.cz> 
 *             
 *  LICENSE TERMS
 *               
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *  3. Neither the name of the Company nor the names of its contributors
 *  may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *                                              
 *  This software is provided ``as is'', and any express or implied
 *  warranties, including, but not limited to, the implied warranties of
 *  merchantability and fitness for a particular purpose are disclaimed.
 *  In no event shall the company or contributors be liable for any
 *  direct, indirect, incidental, special, exemplary, or consequential
 *  damages (including, but not limited to, procurement of substitute
 *  goods or services; loss of use, data, or profits; or business
 *  interruption) however caused and on any theory of liability, whether
 *  in contract, strict liability, or tort (including negligence or
 *  otherwise) arising in any way out of the use of this software, even
 *  if advised of the possibility of such damage.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <libgen.h>
#include <libfasta/fasta.h>
#include <libfasta/trans.h>
#include "libtriplex.h"
#include "gff3.h"
#include "debug.h"

/* default parameters */
#define STEPS           10
#define MIN_SCORE       3
#define MIN_LEN         6
#define MIN_LOOP        3
#define MAX_LOOP        10
#define TRI_TYPE        0
#define P_VALUE         2.0

/* Acceptable command line arguments */
#define ARGUMENTS       "f:o:s:l:i:e:t:p:vhj:"

static void usage(char *me)
{
        printf("Usage: %s [OPTIONS] -f <file.fa> -o <file.gff>\n\n", basename(me));
        printf("\t-f s      Input file\n");
        printf("\t-o s      Output file\n");
        printf("\t-s num    Number of steps (default=%d)\n", STEPS);
        printf("\t-e num    Minimal score treshold (default=%d)\n",MIN_SCORE);
        printf("\t-l num    Minimal triplex length (default=%d)\n", MIN_LEN);
        printf("\t-i num    Minimal triplex loop (default=%d)\n", MIN_LOOP);
        printf("\t-j num    Maximal triplex loop (default=%d)\n", MAX_LOOP);
        printf("\t-t num    Triplex type (0..7) (default=%d)\n", TRI_TYPE);
        printf("\t-p num    Acceptable P-value (default=%1.2f)\n", P_VALUE);
        printf("\t-a num    Lambda for type 0,1,4,5 (default=%1.2f)\n", LAMBDA[0]);
        printf("\t-b num    Lambda for type 2,3,6,7 (default=%1.2f)\n", LAMBDA[2]);
        printf("\t-c num    Mi for type 0,1,4,5 (default=%1.2f)\n", MI[0]);
        printf("\t-d num    Mi for type 2,3,6,7 (default=%1.2f)\n", MI[2]);
        printf("\t-v        Verbose output\n");
        printf("\t-h        Show this text\n");
	printf("\n");
}

/* --------------------------------------------------------------------------
 *                               Main Program
 * --------------------------------------------------------------------------
 */
int main(int argc, char *argv[]) {
	atrans_t    *tr;
	FASTA       *fa;
	FASTA_rec_t  farec_dst, *farec;
	gff3_t      *gff;

	char *arg0 = argv[0];
	char *in_path = NULL, *out_path = NULL;

	tsearch_params_t params;

	int verbose_flag = 0;
	int opt;
	int ch;

	params.steps        = STEPS;
	params.min_length   = MIN_LEN;
	params.min_score    = MIN_SCORE;
	params.triplex_type = TRI_TYPE;
	params.min_loop     = MIN_LOOP;
	params.max_loop     = MAX_LOOP;
	params.p_val        = P_VALUE;

        /* ---------------------------------------------------------------- */
        /*                   Test of input parameters                       */
        /* ---------------------------------------------------------------- */

        while ((opt = getopt(argc, argv, ARGUMENTS)) != -1) {
                switch (opt) {
                case 'f':
                        in_path  = strdup(optarg);
                        break;
		case 'o':
			out_path = strdup(optarg);
			break;
                case 's':
                        params.steps = atoi(optarg);
                        break;
                case 'e':
                        params.min_score = atoi(optarg);
                        break;
                case 'l':
                        params.min_length = atoi(optarg);
                        break;
                case 'i':
                        params.min_loop = atoi(optarg);
                        break;
		case 'j':
			params.max_loop = atoi(optarg);
			break;
                case 't':
                        params.triplex_type = atoi(optarg);
                        break;
		case 'p':
			params.p_val = atof(optarg);
			break;
                case 'a':
                        LAMBDA[0] = LAMBDA[1] = LAMBDA[4] = LAMBDA[5] = atof(optarg);
                        break;

                case 'b':
                        LAMBDA[2] = LAMBDA[3] = LAMBDA[6] = LAMBDA[7] = atof(optarg);
                        break;

                case 'c':
                        MI[0] = MI[1] = MI[4] = MI[5] = atof(optarg);
                        break;

                case 'd':
                        MI[2] = MI[3] = MI[6] = MI[7] = atof(optarg);
                        break;
                case 'v':
                        verbose_flag = 1;
                        break;
                case 'h':
                        usage(arg0);
                        return 0;
                default:
                        printf("ERROR: unknown argument -%c\n", optopt);
                	return 1;
                }
        }

        argc -= optind;
        argv += optind;

        if (argc != 0) {
                printf("ERROR: stray arguments\n");
                return 1;
	}

        if (in_path == NULL || out_path == NULL) {
                usage(arg0);
                return 0;
        }

        if ((params.triplex_type > 7) || (params.triplex_type < 0)) {
                printf("ERROR: triplex type is out of range\n");
                return 1;
        }

	/*
	 * Init GFF 3 container
	 */
	gff = gff3_new(out_path);

	if (gff == NULL) {
		fprintf(stderr, "ERROR: Failed to open the output file: %s\n", out_path);
		return (1);
	}

	/*
	 * Setup alphabet translation
	 */
	tr = atrans_new(8, 8, 0, 0);
	tr->tr_letter_s2d['A'] = tr->tr_letter_s2d['a'] = 0;
	tr->tr_letter_s2d['C'] = tr->tr_letter_s2d['c'] = 1;
	tr->tr_letter_s2d['G'] = tr->tr_letter_s2d['g'] = 2;
	tr->tr_letter_s2d['T'] = tr->tr_letter_s2d['t'] = 3;

	/*
	 * Open the FASTA file
	 */
	fa = fasta_open(in_path, FASTA_READ|FASTA_ONDEMSEQ, tr);

	if (fa == NULL) {
		fprintf(stderr, "ERROR: Failed to open the input file: %s\n", in_path);
		atrans_free(tr);

		return (1);
	}

	while ((farec = fasta_read(fa, &farec_dst,
				   FASTA_INMEMSEQ|FASTA_CSTRSEQ, NULL)) != NULL)
	{
		_D("Searching in %p (%s), seq_len=%zu\n", farec, farec->rec_id, farec->seq_len);

		if (triplex_search(gff, farec->rec_id, farec->seq_mem, farec->seq_len, &params) < 0)
		{
			fprintf(stderr, "ERROR: triplex_search(%s) failed\n", farec->rec_id);
		}

		fasta_rec_free(farec);
	}

	/*
	 * Write output data
	 */
	gff3_write(gff);
	gff3_free(gff);

	/*
	 * Free resources & exit
	 */
	fasta_close(fa);
	atrans_free(tr);

        return 0;
}
